from pathlib import Path
import gmsh


def run_geo(gmsh_geo_file):

    # Initialize Gmsh
    gmsh.initialize()

    # Run the .geo file
    gmsh.open(gmsh_geo_file)
    gmsh.write("mesh.msh")

    # Finalize Gmsh
    gmsh.finalize()

    file = Path("mesh.msh")
    with file.open("rt") as fh:
        string = fh.read()

    return {"gmsh_mesh_str": string}



import numpy as np

ROOT_EPS = np.sqrt(np.finfo(np.float64).eps)


def define_perturbations(trial_params):
    return np.abs(trial_params) * 0.01
    # return ROOT_EPS * (np.abs(trial_params) + 1)


def get_new_model_parameters(trial_params, perturbations):
    """Get sets of parameters at which to evaluate the model.

    Each row is a separate set of parameters, with the first row being the unperturbed
    parameters.
    """
    p_perturbed = np.tile(trial_params, (len(trial_params), 1)) + np.diag(perturbations)
    return np.vstack((trial_params, p_perturbed))


def perturb_parameters(p_0, p_i, p_scales):
    # print(f"perturb_parameters: {p_0[:]=!r}")

    if p_i is not None:
        # already scaled
        # print(f"perturb_parameters: {p_i[:]=!r}")
        # print(f"perturb_parameters: {p_scales[:]=!r}")
        trial_params = np.asarray(p_i)
    else:
        # first iteration; need to scale
        p_scales = p_0
        trial_params = np.ones(len(p_0))

    p_delta = define_perturbations(trial_params)
    # print(f"perturb_parameters: {p_delta=!r}")

    p_perturbed = get_new_model_parameters(trial_params, p_delta)
    # print(f"perturb_parameters: {trial_params=!r}")
    # print(f"perturb_parameters: {p_perturbed=!r}")
    return {
        "p_scales": np.asarray(p_scales),
        "p_perturbed": p_perturbed,
        "perturbations": p_delta,
        "p_i": trial_params,
    }



from typing import Callable, Tuple

import numpy as np
from numpy.typing import NDArray
from numpy.polynomial.polynomial import Polynomial
from scipy.interpolate import lagrange
from scipy.stats import norm


def _to_physical_space(dist, variates):
    """Convert random variables defined in standard normal space to some other
    continuous distribution space."""
    return dist.ppf(norm.cdf(variates, loc=0, scale=1))


def to_physical_space(dists, variates):
    """Convert random variables sampled from the standard normal space to the specified
    distribution spaces.

    Parameters
    ----------
    dists
        List of SciPy frozen distribution instances
    variates
        Standard-normally distributed random variates. Outer dimension must
        match length of `dists` list.
    """
    funcs = [_lookup.get(i.dist.name, ("", _to_physical_space))[1] for i in dists]
    assert len(funcs) == len(variates)
    inputs_t = np.empty_like(variates)
    for idx, (func_i, ins_i) in enumerate(zip(funcs, variates)):
        inputs_t[idx] = func_i(dists[idx], ins_i)
    return inputs_t


_lookup = {}


def get_coarse_model_func(model) -> Tuple[Callable, np.ndarray, np.ndarray]:
    # interpolate at 10 points:
    x_interp = np.linspace(-np.pi, np.pi, 10)
    y_interp = model(x_interp)

    # interpolate points to a Lagrange polynomial:
    poly = lagrange(x_interp, y_interp)
    model_coarse = Polynomial(poly.coef[::-1])

    return model_coarse, x_interp, y_interp


def model(x):
    return np.tanh(x)


def system_analysis_toy_model_DA_coarse(
    x: NDArray, x_0: float, stddev: float, y_star: float
):
    """`x` is within the failure domain if the return is greater than zero."""
    x = x[:]  # convert to numpy array

    input_dist = norm(loc=x_0, scale=stddev)
    x_t = to_physical_space([input_dist], x)

    model_coarse, _, _ = get_coarse_model_func(model)
    g_i = np.squeeze(y_star - model_coarse(x_t)).item()

    return {"g": g_i}



import numpy as np


def get_stress_yy_at_point(FE_response, point, time_inc):
    dist = np.linalg.norm(FE_response["node_coordinates"][0] - point, axis=1)
    closest_idx = np.argmin(dist)
    return FE_response["stress_y"][time_inc, closest_idx, 1]


def calculate_plate_SCF_2D(FE_response, plate_width, hole_diameter, plate_diff):

    hole_node = np.array([hole_diameter / 2, 0.0, 0.0])
    far_node = np.array([plate_width / 4, plate_width / 2 + plate_diff / 4, 0.0])

    print(f"hole_node: {hole_node!r}")
    print(f"far_node: {far_node!r}")

    hole_node_stress = get_stress_yy_at_point(FE_response, hole_node, -1)
    far_node_stress = get_stress_yy_at_point(FE_response, far_node, -1)

    print(f"hole_node: {hole_node!r}")
    print(f"far_node: {far_node!r}")

    # print(f"{hole_node_stress=!r}")
    # print(f"{far_node_stress=!r}")

    stress_CF = (hole_node_stress / far_node_stress).item()

    print(f"stress_CF: {stress_CF!r}")

    # find the expected SCF, correcting for the finite width:
    # width_ratio = hole_diameter / plate_width
    # k_t = 3 - 3.14 * width_ratio + 3.667 * width_ratio**2 - 1.527 * width_ratio**3
    # expected_SCF = k_t / (1 - width_ratio)

    # print(f"calculate_plate_SCF_2D: {width_ratio=!r}")
    # print(f"calculate_plate_SCF_2D: {k_t=!r}")
    # print(f"calculate_plate_SCF_2D: {expected_SCF=!r}")
    # print(f"calculate_plate_SCF_2D: {stress_CF=!r}")

    return {"SCF": stress_CF}



import numpy as np


def model_exp(x, *p):
    # print(f"model_exp: {p}")
    a, b, c = p
    return a * np.exp(b * x) + c


def system_analysis_toy_model(x, p_perturbed, index, p_scales):
    # print(f"system_analysis_toy_model: {x=!r}")
    # print(f"system_analysis_toy_model: {p_perturbed[index]=!r}")
    # print(f"system_analysis_toy_model: {p_scales[:]=!r}")
    # print(f"system_analysis_toy_model: {index=!r}")
    return {
        "system_response": model_exp(
            x, *(i * j for i, j in zip(p_scales, p_perturbed[index]))
        )
    }



import numpy as np


def initialise_markov_chains_ACS(
    chain_index,
    chain_seeds,
    chain_g,
    all_g,
    all_x,
    num_chains_per_update,
    all_accept,
):

    # we pass all iterations of chain_index so we can retrieve the batch index
    # TODO: ideally expose loop indices as env vars

    chain_index_latest_iter = sorted(
        chain_index.items(), key=lambda x: int(x[0].split("_")[1])
    )[-1][1]
    batch_idx = chain_index_latest_iter["loop_idx"]["proposal_update"]
    chain_index = chain_index_latest_iter["value"]

    # chain index within the level (accouting for multiple batches):
    chain_index_level = batch_idx * num_chains_per_update + chain_index

    x = chain_seeds[chain_index_level]
    g = chain_g[chain_index_level]

    if all_g is None:
        all_g = np.array([g])
        all_x = np.array(x)[None]
        all_accept = np.array([], dtype=bool)
    else:
        all_g = np.append(all_g[chain_index], np.array([g]))
        all_x = np.vstack([all_x[chain_index], np.array([x])])
        all_accept = all_accept[chain_index]

    return {"x": x, "g": g, "all_x": all_x, "all_g": all_g, "all_accept": all_accept}



def perform_inference(
    surrogate,
    Y_test,
    test_index,
    test_error,
    draws,
    tune,
    chains,
    cores,
    fitted_x,
    use_std,
):
    data = Y_test[test_index]
    error = test_error * data

    inf_dat = surrogate.perform_inference(
        Y_actual=data,
        Y_error=error,
        draws=draws,
        tune=tune,
        chains=chains,
        cores=cores,
        return_inferencedata=True,
        initval=fitted_x,
        use_std=use_std,
    )

    return {"inference_data": inf_dat}



from matflow.param_classes import SingleCrystalParameters


def test_perturbation():
    params = SingleCrystalParameters(
        phases={"Al": {"xi_0_sl": [30e6]}},
        perturbations={"path": ["Al", "xi_0_sl", 0], "multiplicative": 1.05},
    )
    assert params.phases == {"Al": {"xi_0_sl": [30e6 * 1.05]}}


def test_perturbation_list():
    params = SingleCrystalParameters(
        phases={"Al": {"xi_0_sl": [30e6]}},
        perturbations=[
            {"path": ["Al", "xi_0_sl", 0], "multiplicative": 1.05},
            {"path": ["Al", "xi_0_sl", 0], "multiplicative": 1.05},
        ],
    )
    assert params.phases == {"Al": {"xi_0_sl": [30e6 * 1.05 * 1.05]}}


def test_perturbation_init_to_dict():
    params = SingleCrystalParameters(
        phases={"Al": {"xi_0_sl": [30e6]}},
        perturbations={"path": ["Al", "xi_0_sl", 0], "multiplicative": 1.05},
    )
    SingleCrystalParameters(**params.to_dict())



from __future__ import annotations

from matflow.param_classes.moose import MooseInputDeck


def write_input(
    path,
    input_deck: MooseInputDeck | None = None,
    input_deck_variables=None,
    elasticity=None,
    _exodus_generation_input_deck: MooseInputDeck | None = None,
):

    if _exodus_generation_input_deck is not None:
        # this is just for generating an Exodus file from a Gmsh .msh file
        # for easier visualisation:
        input_deck = _exodus_generation_input_deck

    if not isinstance(input_deck, MooseInputDeck):
        input_deck = MooseInputDeck(**input_deck)

    if input_deck_variables:
        input_deck.add_variables(input_deck_variables)
    if elasticity:
        input_deck.update(
            {("Materials", "elasticity", k): v for k, v in elasticity.items()}
        )
    input_deck.to_file(path)



import numpy as np

from matflow.param_classes.utils import masked_array_from_list


def test_masked_array_from_list_1D():
    arr_lst = [1, 2, 3]
    assert np.allclose(masked_array_from_list(arr_lst), np.array(arr_lst))


def test_masked_array_from_list_2D():
    arr_lst = [[1, 2, 3], [4, 5, 6]]
    assert np.allclose(masked_array_from_list(arr_lst), np.array(arr_lst))


def test_masked_array_from_list_1D_masked():
    arr_lst = [1, 2, "x"]
    expected_data = np.array([1, 2, -1])
    expected_mask = np.array([0, 0, 1])
    expected = np.ma.masked_array(data=expected_data, mask=expected_mask)
    assert np.allclose(masked_array_from_list(arr_lst), expected)


def test_masked_array_from_list_2D_masked():
    arr_lst = [
        [1, 2, "x"],
        [3, 4, "x"],
        [5, 6, "x"],
    ]
    expected_data = np.array(
        [
            [1, 2, -1],
            [3, 4, -1],
            [5, 6, -1],
        ]
    )
    expected_mask = np.array(
        [
            [0, 0, 1],
            [0, 0, 1],
            [0, 0, 1],
        ]
    )
    expected = np.ma.masked_array(data=expected_data, mask=expected_mask)
    assert np.allclose(masked_array_from_list(arr_lst), expected)



import numpy as np
import sklearn
from matflow.param_classes.surrogate import Surrogate


def build_surrogate(
    X_train,
    Y_train,
    parameter_names,
    n_restarts_optimizer,
    cross_validate,
    scoring,
    normalize_y,
    validate,
    X_test,
    Y_test,
):
    surrogate = Surrogate(X_train, Y_train, parameter_names=parameter_names, scale=True)
    surrogate.build_model(
        n_restarts_optimizer=n_restarts_optimizer,
        cross_validate=cross_validate,
        scoring=scoring,
        normalize_y=normalize_y,
    )

    # validation:
    vld_out = {}
    if validate:
        Y_prediction, Y_error = surrogate.make_prediction(X_test, return_std=True)
        score = sklearn.metrics.r2_score(Y_test, Y_prediction)
        print(f"Validation score: {score}.")
        vld_out = {
            "Y_prediction": Y_prediction,
            "Y_error": Y_error,
            "score": score,
        }

    # versions: once serialised, the model (e.g. GaussianProcessRegressor) may not be
    # correctly deserialised when using different versions of scikit-learn and its
    # dependencies:
    versions = {
        "numpy": np.__version__,
        "sklearn": sklearn.__version__,
    }

    return {"surrogate": surrogate, "validation": vld_out, "versions": versions}



"""
Custom encoders and decoders for parameter data.
"""

from collections import defaultdict

from hpcflow.sdk.core.utils import get_relative_path, get_in_container, set_in_container


def _try_import_arviz():
    """Try to import Arviz, a package for Bayesian model analysis."""
    try:
        import arviz

        return arviz
    except ImportError:
        return False


def _zarr_encode_arviz_inference_data(
    obj,
    type_lookup,
    path,
    root_encoder,
    **kwargs,
):
    """Encode to the Zarr store an InferenceData object from the arviz package."""
    type_lookup["arviz.data.inference_data.InferenceData"].append(path)
    return root_encoder(obj.to_dict(), type_lookup=type_lookup, path=path, **kwargs)[
        "data"
    ]


def _zarr_decode_arviz_inference_data(
    obj,
    type_lookup,
    path,
    **kwargs,
):
    """Decode from the Zarr store an InferenceData object from the arviz package."""
    if arviz := _try_import_arviz():
        for dat_path in type_lookup.get("arviz.data.inference_data.InferenceData", []):
            try:
                rel_path = get_relative_path(dat_path, path)
            except ValueError:
                continue

            if rel_path:
                set_in_container(
                    obj, rel_path, arviz.from_dict(**get_in_container(obj, rel_path))
                )
            else:
                obj = arviz.from_dict(**obj)

    return obj


def get_encoders():
    """Get additional app defined encoders."""
    encoders = defaultdict(dict)
    if arviz := _try_import_arviz():
        encoders["zarr"][
            arviz.data.inference_data.InferenceData
        ] = _zarr_encode_arviz_inference_data

    return encoders


def get_decoders():
    """Get additional app defined decoders."""
    decoders = defaultdict(dict)
    if _try_import_arviz():
        decoders["zarr"][
            "arviz.data.inference_data.InferenceData"
        ] = _zarr_decode_arviz_inference_data

    return decoders



def refine_plate_mesh(
    SCF, hole_sect_nodes, plate_radial_nodes, plate_diff_nodes, plate_thick_layers
):
    scale = 1.2
    hole_sect_nodes = int(hole_sect_nodes * scale)
    plate_radial_nodes = int(plate_radial_nodes * scale)
    plate_diff_nodes = int(plate_diff_nodes * scale)
    plate_thick_layers = plate_thick_layers + 1

    # should be odd:
    if hole_sect_nodes % 2 == 0:
        hole_sect_nodes += 1
    if plate_diff_nodes % 2 == 0:
        plate_diff_nodes += 1

    return {
        "hole_sect_nodes": hole_sect_nodes,
        "plate_radial_nodes": plate_radial_nodes,
        "plate_diff_nodes": plate_diff_nodes,
        "plate_thick_layers": plate_thick_layers,
    }



import numpy as np
import pyvista as pv
from pyvista.core.utilities import VtkErrorCatcher


SHAPES = {
    "stress_x": (3,),
    "stress_y": (3,),
    "stress_z": (3,),
    "disp_": (3,),
}


def parse_exodus(moose_exodus_file, FE_response_data):

    with VtkErrorCatcher(send_to_logging=False):
        # note: this context manager reduces the output to stderr from VTK, but we still
        # currently get some output, and I can't find a way to silence it.

        path = moose_exodus_file[0]  # list because the file is defined as a regex

        reader = pv.ExodusIIReader(path)
        times = reader.time_values

        reader.set_active_time_point(0)
        mesh = reader.read()
        dataset = mesh.combine()
        num_nodes = dataset.points.shape[0]
        num_incs = len(times)

        FE_response = {
            "times": times,
            "node_coordinates": np.empty((len(times), num_nodes, 3)),
        }
        for key in FE_response_data:
            new_shape = (len(times), num_nodes, *SHAPES.get(key, tuple()))
            FE_response[key] = np.empty(new_shape)

        for time_idx in range(num_incs):
            if time_idx > 0:
                reader.set_active_time_point(time_idx)
                mesh = reader.read()
                dataset = mesh.combine()
            FE_response["node_coordinates"][time_idx] = dataset.points
            for key in FE_response_data:
                FE_response[key][time_idx] = dataset.point_data[key]

    return FE_response



"""
Runs the command line interface.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import click
from hpcflow.sdk.submission.shells import DEFAULT_SHELL_NAMES


import matflow as mf
from matflow import cli
from matflow.environments import (
    env_configure_moose,
    env_configure_python,
    env_configure_python_all,
    env_configure_dream3d,
    env_configure_matlab,
    env_configure_damask,
)


if TYPE_CHECKING:
    from hpcflow.sdk.app import BaseApp


def add_to_env_setup_CLI(app: BaseApp) -> click.Group:
    """Generate the CLI for configuring MatFlow environments."""

    shell = DEFAULT_SHELL_NAMES[os.name]

    @click.command()
    @click.option(
        "-p",
        "--pipeline-runner-path",
        required=True,
        type=click.Path(exists=True, file_okay=True),
        help="Absolute path to the pipeline runner executable.",
    )
    @click.option(
        "--use-current/--no-use-current",
        is_flag=True,
        default=True,
        help=(
            "Use the currently active conda-like or Python virtual environment to add a "
            "`python_script` executable to the environment."
        ),
    )
    def dream3d(pipeline_runner_path: Path, use_current: bool):
        """Configure the Dream3D environment.

        The path to the PipelineRunner executable must be provided.

        """
        env = env_configure_dream3d(
            shell,
            pipeline_runner_path=pipeline_runner_path,
            use_current=use_current,
        )
        app.save_env(env)

    @click.command()
    @click.option(
        "-n",
        "--name",
        multiple=True,
        help=(
            "In addition to the `python_env` set up these other named environments "
            '(suffixed by "_env"), also with a `python_script` executable.'
        ),
    )
    @click.option(
        "--use-current/--no-use-current",
        is_flag=True,
        default=True,
        help=(
            "Use the currently active conda-like or Python virtual environment to add a "
            "`python_script` executable to the environment."
        ),
    )
    def python(name: list[str], use_current: bool):
        """Configure environments with `python_script` executables."""
        envs = env_configure_python(
            shell,
            names=name,
            use_current=use_current,
        )
        for env in envs:
            app.logger.debug(f"Saving 'python' environment: {env.name!r}.")
            app.save_env(env)

    @click.command()
    @click.option(
        "--use-current/--no-use-current",
        is_flag=True,
        default=True,
        help=(
            "Use the currently active conda-like or Python virtual environment to add a "
            "`python_script` executable to the environment."
        ),
    )
    def python_all(use_current: bool):
        """Configure all Python environments with the same setup."""
        envs = env_configure_python_all(shell, use_current=use_current)
        for env in envs:
            app.logger.debug(f"Saving 'python' environment: {env.name!r}.")
            app.save_env(env)

    @click.command()
    @click.option(
        "--path",
        type=click.Path(exists=True, dir_okay=True),
        help="Absolute path to the MATLAB installation directory.",
    )
    @click.option(
        "--runtime-path",
        type=click.Path(exists=True, dir_okay=True),
        help="Absolute path to the MATLAB runtime directory.",
    )
    @click.option(
        "--mtex-path",
        type=click.Path(exists=True, dir_okay=True),
        help="Absolute path to the MTEX installation directory.",
    )
    def matlab(path: Path, runtime_path: Path, mtex_path: Path):
        """Configure the MATLAB environment for running/compiling MTEX scripts."""
        env = env_configure_matlab(
            shell,
            matlab_path=path,
            matlab_runtime_path=runtime_path,
            mtex_path=mtex_path,
        )
        app.save_env(env)

    @click.command()
    @click.option(
        "--docker-image",
        help=(
            "Name of the docker image to use to run DAMASK_grid; or the name of the "
            "image expected within the archive."
        ),
    )
    @click.option(
        "--docker-archive",
        help="Name of the docker archive to load the image from.",
    )
    @click.option(
        "--singularity-archive",
        help="Name of the docker archive to build a sif file from.",
    )
    @click.option(
        "--singularity-sif",
        help="Name of the singularity sif file to use.",
    )
    @click.option(
        "--docker-exe",
        default="docker",
        help="Docker executable file.",
    )
    @click.option(
        "--singularity-exe",
        default="singularity",
        help="Singularity executable file.",
    )
    def damask(
        docker_image: str | None,
        docker_archive: str | Path | None,
        singularity_archive: str | Path | None,
        singularity_sif: str | Path | None,
        docker_exe: str | Path,
        singularity_exe: str | Path,
    ):
        """Configure the DAMASK environment.

        In particular the `damask_grid` executable.

        """
        env = env_configure_damask(
            shell,
            docker_image=docker_image,
            docker_archive=docker_archive,
            docker_exe=docker_exe,
            singularity_archive=singularity_archive,
            singularity_sif=singularity_sif,
            singularity_exe=singularity_exe,
        )
        app.save_env(env)

    @click.command()
    @click.option(
        "--docker-image",
        help=(
            "Name of the docker image to use to run DAMASK_grid; or the name of the "
            "image expected within the archive."
        ),
    )
    @click.option(
        "--docker-archive",
        help="Name of the docker archive to load the image from.",
    )
    @click.option(
        "--singularity-archive",
        help="Name of the docker archive to build a sif file from.",
    )
    @click.option(
        "--singularity-sif",
        help="Name of the singularity sif file to use.",
    )
    @click.option(
        "--docker-exe",
        default="docker",
        help="Docker executable file.",
    )
    @click.option(
        "--singularity-exe",
        default="singularity",
        help="Singularity executable file.",
    )
    def moose(
        docker_image: str | None,
        docker_archive: str | Path | None,
        singularity_archive: str | Path | None,
        singularity_sif: str | Path | None,
        docker_exe: str | Path,
        singularity_exe: str | Path,
    ):
        """Configure the MOOSE (proteus) environment."""
        env = env_configure_moose(
            shell,
            docker_image=docker_image,
            docker_archive=docker_archive,
            docker_exe=docker_exe,
            singularity_archive=singularity_archive,
            singularity_sif=singularity_sif,
            singularity_exe=singularity_exe,
        )
        app.save_env(env)

    app.env_setup_CLI.add_command(dream3d)
    app.env_setup_CLI.add_command(python)
    app.env_setup_CLI.add_command(python_all)
    app.env_setup_CLI.add_command(matlab)
    app.env_setup_CLI.add_command(damask)
    app.env_setup_CLI.add_command(moose)


add_to_env_setup_CLI(mf)

if __name__ == "__main__":
    cli()



import numpy as np


def solve_new_trial_parameters(
    damping_factors, jacobian, y_ref, y_unperturbed, trial_params
):
    # print(f"solve_new_trial_parameters: {damping_factors=!r}")
    # print(f"solve_new_trial_parameters: {jacobian=!r}")
    # print(f"solve_new_trial_parameters: {y_ref=!r}")
    # print(f"solve_new_trial_parameters: {y_unperturbed=!r}")
    # print(f"solve_new_trial_parameters: {trial_params=!r}")
    y_diff = y_ref - y_unperturbed
    b = np.hstack([y_diff, np.zeros(jacobian.shape[1])])

    kappas = []
    errors = []
    for damping in damping_factors:

        A = np.vstack([jacobian, np.sqrt(damping) * np.eye(jacobian.shape[1])])
        kappa, *_ = np.linalg.lstsq(A, b, rcond=None)
        kappas.append(kappa)
        errors.append(np.sum(np.abs(y_diff - jacobian @ kappa)))

    best_idx = np.argmin(errors)
    best_kappa = kappas[best_idx]
    new_damping = damping_factors[best_idx]

    return (trial_params + best_kappa, new_damping)


def approximate_jacobian(perturbations, y_unperturbed, y_perturbed):
    """
    Parameters
    ----------
    perturbations
        Perturbation for each parameter
    y_unperturbed
        Vector of the unperturbed model response.
    y_perturbed:
        Array of column vectors where each column corresponds to the model response where
        one parameter has been perturbed.

    """
    # print(f"approximate_jacobian: {perturbations[:]=!r}")
    # print(f"approximate_jacobian: {y_unperturbed[:]=!r}")
    # print(f"approximate_jacobian: {y_perturbed[:]=!r}")
    return (y_perturbed - y_unperturbed[:, None]) / perturbations


def get_trial_damping_factors(current_damping=None):
    base = np.array([0.1, 0.2, 0.5, 1.0, 2.0, 5.0, 10.0])
    return base * current_damping if current_damping is not None else base


def lm_fit(
    system_response,
    reference_response,
    perturbations,
    damping,
    p_i,
    p_scales,
    stop_tol,
):
    # print(f"lm_fit: {stop_tol=!r}")
    # print(f"lm_fit: {perturbations=!r}")
    # print(f"lm_fit: {p_scales=!r}")
    # print(f"lm_fit: {damping=!r}")
    # print(f"lm_fit: {p_i=!r}")
    # print(f"lm_fit: {reference_response=!r}")
    # print(f"lm_fit: {system_response=!r}")

    y_unpert = system_response[0][:]
    y_pert = np.asarray([resp[:] for resp in system_response[1:]])

    jac_approx = approximate_jacobian(perturbations, y_unpert, y_pert.T)
    # print(f"lm_fit: {jac_approx=!r}")
    p_i_next, new_damping = solve_new_trial_parameters(
        damping_factors=get_trial_damping_factors(damping),
        jacobian=jac_approx,
        y_ref=reference_response,
        y_unperturbed=y_unpert,
        trial_params=p_i,
    )
    print(f"lm_fit: {p_i[:]=!r}")
    print(f"lm_fit: {p_i_next[:]=!r}")
    print(f"lm_fit: {p_scales[:]=!r}")
    p_scales = p_scales[0] if isinstance(p_scales, list) else p_scales
    fit_params = p_scales * p_i_next
    is_converged = bool(np.linalg.norm(p_i_next - p_i) < stop_tol)
    # print(f"lm_fit: (2) {p_scales=!r}")
    print(f"lm_fit: {fit_params=!r}; {new_damping=!r} {is_converged=!r}")
    return {
        "p_i": p_i_next,
        "fit_parameters": fit_params,
        "damping": new_damping,
        "is_converged": is_converged,
    }



def fit_surrogate(surrogate, Y_test, test_index, test_error, iters, n, use_std):

    data = Y_test[test_index]
    error = test_error * data

    res = surrogate.fit(
        Y_actual=data,
        Y_error=error,
        iters=iters,
        n=n,
        use_std=use_std,
    )
    print(f"Fitted parameters: {res.x}")
    return {"fitted_x": res.x, "optimize_result": dict(res)}



from textwrap import dedent

from matflow.param_classes.moose import MooseInputDeck


def test_round_trip_no_comments():
    string = dedent(
        """\
        [Mesh]
            [generated]
                type = GeneratedMeshGenerator
                dim = 2
                nx = 10
                ny = 5
                xmax = 0.1
                ymax = 0.05
                elem_type = QUAD4
            []
        []
        [Variables]
            [temperature]
                initial_condition = 20.0
            []
        []
        [Kernels]
            [heat_conduction]
                type = HeatConduction
                variable = temperature
            []
        []
        [Materials]
            [copper_thermal]
                type = HeatConductionMaterial
                thermal_conductivity = 384.0
                specific_heat = 406.0
            []
            [copper_density]
                type = GenericConstantMaterial
                prop_names = density
                prop_values = 8829.0
            []
        []
        [BCs]
            [heat_flux_out]
                type = ConvectiveHeatFluxBC
                variable = temperature
                boundary = left
                T_infinity = 20.0
                heat_transfer_coefficient = 125000.0
            []
            [heat_flux_in]
                type = NeumannBC
                variable = temperature
                boundary = right
                value = 500000.0
            []
        []
        [Executioner]
            type = Steady
        []
        [Postprocessors]
            [max_temp]
                type = NodalExtremeValue
                variable = temperature
            []
            [avg_temp]
                type = AverageNodalVariableValue
                variable = temperature
            []
        []
        [Outputs]
            exodus = True
        []
    """
    )
    inp = MooseInputDeck.from_string(string)
    assert inp.block_dat["Variables"]["temperature"]["initial_condition"] == "20.0"
    assert inp.block_dat["Postprocessors"]["max_temp"]["variable"] == "temperature"
    assert inp.to_string() == string



import numpy as np


def process_moose_response(FE_response):
    # node_coords = FE_response["node_coordinates"][0]  # at time step zero
    node_disp = FE_response["disp_"][-1]  # at final time step
    node_disp_xy_flat = node_disp[:, [0, 1]].flatten()

    # print(f"process_moose_response: {node_coords.shape=!r}")
    # print(f"process_moose_response: {node_disp.shape=!r}")
    # print(f"process_moose_response: {node_disp_xy_flat.shape=!r}")
    # print(f"process_moose_response: {node_disp_xy_flat=!r}")
    return {"system_response": node_disp_xy_flat}



from __future__ import annotations
import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any

from damask import __version__ as damask_version
from damask_parse.writers import write_geom, write_load_case, write_material
from damask_parse import write_numerics

if TYPE_CHECKING:
    from matflow.param_classes.load import LoadCase


def write_input_files(
    path: Path | str,
    volume_element: dict,
    load_case: LoadCase,
    damask_solver: dict[str, str],
    homogenization: dict,
    damask_phases: dict,
    single_crystal_parameters: dict | None,
    damask_numerics: dict | None,
    initial_conditions: dict[str, Any] | None,
):
    """
    Write all the input files to DAMASK.

    Parameters
    ----------
    path
        Path to the geometry file to write.
    volume_element
        The volume element descriptor.
    load_case
        The loading case plan.
    damask_solver
        Configuration of the solver to use.
    homogenization
        Homogenisation control.
    damask_phases
        Phase information map.
    single_crystal_parameters
        Overrides for information about single crystals in the phase information map.
    damask_numerics
        Optional dict of key-value pairs to write into the DAMASK numerics control file.
        https://damask-multiphysics.org/documentation/file_formats/numerics.html
    """
    geom_path = Path(path)
    _write_geom(geom_path, volume_element, initial_conditions=initial_conditions)
    _write_load(load_case, damask_solver)
    _write_material(
        volume_element,
        homogenization,
        damask_phases,
        single_crystal_parameters,
    )
    if damask_numerics:
        _write_numerics(damask_numerics)


def _write_geom(
    path: Path, volume_element: dict, initial_conditions: dict[str, Any] | None
):
    write_geom(
        dir_path=path.parent,
        volume_element=volume_element,
        name=path.name,
        initial_conditions=initial_conditions,
    )


def _write_load(load_case: LoadCase, damask_solver: dict[str, str]):
    path = Path("load.yaml")
    write_load_case(
        dir_path=path.parent,
        load_cases=load_case.create_damask_loading_plan(),
        solver=damask_solver,
        name=path.name,
        write_2D_arrs=(damask_version != "3.0.0-alpha3"),
    )


def _write_material(
    volume_element: dict,
    homogenization: dict,
    damask_phases: dict,
    single_crystal_parameters: dict | None,
):
    path = Path("material.yaml")
    if single_crystal_parameters is not None:
        # merge single-crystal properties into phases:
        damask_phases = copy.deepcopy(damask_phases)
        for phase_label in damask_phases.keys():
            SC_params_name = damask_phases[phase_label]["mechanical"]["plastic"].pop(
                "single_crystal_parameters", None
            )
            if SC_params_name:
                SC_params = single_crystal_parameters[SC_params_name]
                damask_phases[phase_label]["mechanical"]["plastic"].update(**SC_params)

    write_material(
        dir_path=path.parent,
        homog_schemes=homogenization,
        phases=damask_phases,
        volume_element=volume_element,
        name=path.name,
    )


def _write_numerics(damask_numerics: dict):
    path = Path("numerics.yaml")
    write_numerics(dir_path=path.parent, numerics=damask_numerics, name=path.name)



"""
The moose input deck.
"""

from __future__ import annotations
import re

import logging
from pathlib import Path
from typing import Any
from typing_extensions import ClassVar

from hpcflow.sdk.core.parameters import ParameterValue
from hpcflow.sdk.core.utils import set_in_container

logger = logging.getLogger(__name__)


class MooseBlock:
    """A (recursive) moose block within the input deck file."""

    TAB = "    "

    def __init__(
        self,
        name,
        collection,
        root=False,
        variables=None,
        exclude: Sequence[str] | None = None,
    ):
        self.name = name
        self.root = root
        self.attributes = {}
        self.blocks = []
        self.exclude = tuple(exclude or tuple())
        self.variables = {k: v for k, v in (variables or {}).items() if k not in exclude}

        for key, val in collection.items():
            if key in self.exclude:
                continue
            if isinstance(val, dict):
                self.blocks.append(MooseBlock(key, val))
                continue
            self.attributes[key] = val

    def __eq__(self):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.name == other.name
            and self.attributes == other.attributes
            and self.blocks == other.blocks
            and self.variables == other.variables
        )

    def __str__(self) -> str:
        if self.root:
            tab = ""
            txt = ""
            for key, val in self.variables.items():
                txt += f"{key} = {val}\n"
        else:
            tab = self.TAB
            txt = f"[{self.name}]\n"

        for key, val in self.attributes.items():
            txt += f"{tab}{key} = {val}\n"
        for block in self.blocks:
            txt += tab
            txt += str(block).replace("\n", f"\n{tab}")
            txt = txt[: len(txt) - len(tab)]
        if not self.root:
            txt += "[]\n"
        return txt

    def to_file(self, path: Path):
        with path.open("wt", newline="\n") as f:
            f.write(self.__str__())


class MooseInputDeck(ParameterValue):
    """Moose input parametrisation.

    Notes
    -----
    We don't attempt to parse value types (e.g. floats, ints); everything stays as a
    string.
    """

    _typ: ClassVar[str] = "input_deck"

    def __init__(self, variables: dict | None = None, **block_dat) -> None:
        self.block_dat = block_dat
        self.variables = variables or {}

    def update(self, upd: dict[str | tuple[str], Any]):
        """Update paths within the input deck with the specified values."""
        for path, val in upd.items():
            if isinstance(path, str):
                path_t = tuple(path.split("."))
            else:
                path_t = path
            set_in_container(cont=self.block_dat, path=path_t, value=val)

    def add_variables(self, vars):
        """Add variables to the input deck."""
        self.variables.update(vars)

    def to_string(self):
        """Transform to a string."""
        block = MooseBlock(
            "root",
            self.block_dat,
            root=True,
            variables=self.variables,
            exclude=("__comments__",),
        )
        return str(block)

    def to_file(self, path: str | Path):
        """Write the input file to disk."""
        Path(path).write_text(self.to_string(), newline="\n")

    @staticmethod
    def parse_from_string(contents: str) -> tuple[dict[str, Any], dict[str, Any]]:
        """Parse the contents of a MOOSE input deck file, returning the both the nested
        block structure, and any top-level variables."""

        variables = {}
        root = {}
        stack = [(None, root)]  # (block_name, dict)

        # patterns for [BlockName], [] (end), or key = value lines
        block_start = re.compile(r"^\[(.*?)\]$")
        assignment = re.compile(r"^(\w+)\s*=\s*(.+)$")

        for line in contents.splitlines():
            line = line.strip()
            if not line:
                continue

            # extract comments
            comment = None
            if "#" in line:
                line, comment = line.split("#", 1)
                line = line.rstrip()
                comment = comment.strip()

            # store full line comments
            if line == "" and comment:
                # Store comment in the current block
                block_dict = stack[-1][1]
                block_dict.setdefault("__comments__", []).append(comment)
                continue

            look_for_vars = len(stack) == 1

            # block end:
            if line == "[]":
                stack.pop()
                continue

            # block start
            blk_match = block_start.match(line)
            if blk_match:
                name = blk_match.group(1).strip()
                new_dict = {}
                _, parent_dict = stack[-1]

                # insert into parent dict:
                parent_dict[name] = new_dict

                # inline comments
                if comment:
                    new_dict.setdefault("__comments__", []).append(comment)

                # Push to stack
                stack.append((name, new_dict))
                continue

            # assign values:
            assign_match = assignment.match(line)
            if assign_match:
                key, value = assign_match.group(1), assign_match.group(2).strip()
                if look_for_vars:
                    block_dict = variables
                else:
                    block_dict = stack[-1][1]

                block_dict[key] = value
                if comment:
                    block_dict.setdefault("__comments__", []).append(f"{key}: {comment}")
                continue

            raise ValueError(f"Unrecognized line format: {line!r}.")

        return root, variables

    @classmethod
    def from_string(cls, string, updates: dict[str | tuple[str], Any] | None = None):
        """Generate from a string, like that found in a file."""
        block_dat, variables = cls.parse_from_string(string)
        obj = cls(variables=variables, **block_dat)
        if updates:
            obj.update(updates)
        return obj

    @classmethod
    def from_file(
        cls, path: Path | str, updates: dict[str | tuple[str], Any] | None = None
    ):
        """Generate from an existing input file, optionally with updates to the specified
        paths."""
        return cls.from_string(Path(path).read_text(), updates=updates)



from damask_parse import read_material, read_geom
from damask_parse.utils import validate_orientations, validate_volume_element


def generate_volume_element_from_damask_input_files(
    geom_path, material_path, orientations
):
    geom_dat = read_geom(geom_path)
    material_data = read_material(material_path)
    volume_element = {
        "element_material_idx": geom_dat["element_material_idx"],
        "grid_size": geom_dat["grid_size"],
        "size": geom_dat["size"],
        **material_data["volume_element"],
    }

    if orientations is not None:
        orientations = validate_orientations(orientations)
        num_supplied_ori = orientations["quaternions"].shape[0]
        num_material_ori = volume_element["orientations"]["quaternions"].shape[0]
        if num_supplied_ori != num_material_ori:
            raise ValueError(
                f"Number of orientations supplied {num_supplied_ori} is different to "
                f"number in the material file {num_material_ori}."
            )

        volume_element["orientations"] = orientations

    volume_element = validate_volume_element(volume_element)
    out = {"volume_element": volume_element}
    return out



import numpy as np
import scipy


def collate_CP_data(VE_response, workflow):

    # Path variable names
    paths = (
        workflow.tasks.simulate_VE_loading_damask.template.element_sets[0]
        .multi_path_sequences[0]
        .paths
    )

    # We assume the final section of parameters is the appropriate variable name
    fitting_param_names = [path.split("plastic.")[-1] for path in paths]

    # Determine the scaling used, either linear or logarithmic
    scale = [
        workflow.tasks.simulate_VE_loading_damask.template.element_sets[0]
        .multi_path_sequences[0]
        ._values_method_args["bounds"][path]["scaling"]
        for path in paths
    ]

    # Now determine the bounds of the parameter space
    bounds = [
        workflow.tasks.simulate_VE_loading_damask.template.element_sets[0]
        .multi_path_sequences[0]
        ._values_method_args["bounds"][path]["extent"]
        for path in paths
    ]

    # Now apply the log or linear transformations to the bounds
    bounds = [
        np.log10(bounds[i]) if scale[i] == "log" else bounds[i]
        for i in range(len(bounds))
    ]

    X = np.linspace(0.0, 0.09, 150)
    y_simulation = []
    theta = []

    for VE_idx, VE_resp_i in enumerate(VE_response):

        element = workflow.tasks.simulate_VE_loading_damask.elements[VE_idx]

        try:

            # Read the parameters
            params = [element.get(path) for path in paths]

            # Scale the parameters
            scaled_params = [
                np.log10(params[i]) if scale[i] == "log" else params[i]
                for i in range(len(params))
            ]

            # Extract the stress strain curves
            true_stress_tensor = np.asarray(
                VE_resp_i["phase_data"]["vol_avg_equivalent_stress"]["data"]
            )
            true_strain_tensor = np.asarray(
                VE_resp_i["phase_data"]["vol_avg_equivalent_strain"]["data"]
            )

            stress = true_stress_tensor[:, 0, 0]
            strain = true_strain_tensor[:, 0, 0]

            f_stress = scipy.interpolate.interp1d(x=strain, y=stress)

            # save the outputs
            y_simulation.append(f_stress(X))

            # Save the parameters
            theta.append(scaled_params)

        except Exception as e:
            print("Read Error:", e)
            y_simulation.append(np.zeros_like(X))
            theta.append(scaled_params)

    return {
        "y_simulation": np.array(y_simulation),
        "theta": np.array(theta),
        "parameter_names": fitting_param_names,
    }



from hpcflow.sdk import app as sdk_app
from hpcflow.sdk.config import ConfigOptions

from matflow._version import __version__
from matflow.encoders import get_encoders, get_decoders


# provide access to app attributes:
__getattr__ = sdk_app.get_app_attribute

# ensure docs/help can see dynamically loaded attributes:
__all__ = sdk_app.get_app_module_all()
__dir__ = sdk_app.get_app_module_dir()

# set app-level config options:
config_options = ConfigOptions(
    directory_env_var="MATFLOW_CONFIG_DIR",
    default_directory="~/.matflow",
    default_known_configs_dir="github://hpcflow:matflow-configs@main",
)

# load built in template components:
template_components = sdk_app.App.load_builtin_template_component_data(
    "matflow.data.template_components"
)

# initialise the App object:
app: sdk_app.App = sdk_app.App(
    name="MatFlow",
    version=__version__,
    module=__name__,
    docs_import_conv="mf",
    description="Materials science workflow manager",
    gh_org="hpcflow",
    gh_repo="matflow",
    template_components=template_components,
    scripts_dir="data.scripts",  # relative to root package
    jinja_templates_dir="data.jinja_templates",  # relative to root package
    workflows_dir="data.workflows",  # relative to root package
    config_options=config_options,
    data_manifest_dir="matflow.data.data_manifests",
    data_dir="github://hpcflow:matflow-data@main/data",
    program_dir="github://hpcflow:matflow-data@main/programs",
    docs_url="https://docs.matflow.io/stable",
    encoders=get_encoders,
    decoders=get_decoders,
)  #: |app|

# defer import to allow us to use the app logger in the ParameterValue classes:
from matflow.param_classes import *



import pytest
import numpy as np
import matflow as mf


@pytest.mark.xfail(reason="Requires the Arviz (Bayesian inference) package.")
def test_encode_decode_arviz_inference_data(null_config, tmp_path):

    import arviz

    p1 = np.array([1, 2, 3])
    p2 = (1, 2, 3)
    s1 = mf.TaskSchema(
        "t1",
        inputs=[
            mf.SchemaInput("inference_data"),
            mf.SchemaInput("p1"),
            mf.SchemaInput("p2"),
        ],
    )
    t1 = mf.Task(
        schema=s1,
        inputs={
            "inference_data": {"a": arviz.data.inference_data.InferenceData()},
            "p1": p1,
            "p2": p2,
        },
    )
    wk = mf.Workflow.from_template_data(
        template_name="test",
        tasks=[t1],
        store="zarr",
        path=tmp_path,
    )

    assert isinstance(
        wk.tasks[0].elements[0].inputs.inference_data.value["a"], arviz.InferenceData
    )
    assert np.allclose(wk.tasks[0].elements[0].inputs.p1.value, p1)
    assert wk.tasks[0].elements[0].inputs.p2.value == p2



"""
Functions for configuring MatFlow environments.

"""

from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Literal
import sys

import matflow as mf

from hpcflow.sdk.submission.schedulers.utils import run_cmd

if TYPE_CHECKING:
    from hpcflow.sdk.core.environment import Environment


def __norm_setup(setup: str | list[str] | None = None):
    if setup is None:
        return []
    if isinstance(setup, str):
        return [setup]
    return setup


def __get_py_exe(shell: Literal["bash", "powershell"]) -> str:
    return {
        "bash": sys.executable,
        "powershell": f"& '{sys.executable}'",
    }[shell]


def env_configure_dream3d(
    shell: Literal["bash", "powershell"],
    pipeline_runner_path: str,
    setup_py: str | list[str] | None = None,
    setup_runner: str | list[str] | None = None,
    use_current: bool = True,
) -> Environment:
    """Configure the Dream3D MatFlow environment.

    The Dream3D environment requires:

    1. a `dream_3D_runner` executable that runs the PipelineRunner CLI tool, and
    2. a `python_script` executable that runs the pre and post processing.

    """

    if not Path(pipeline_runner_path).is_file():
        raise ValueError(
            f"The specified Dream3D pipeline runner path is not a file: "
            f"{pipeline_runner_path!r}."
        )

    setup_py = __norm_setup(setup_py)
    setup_runner = __norm_setup(setup_runner)
    RUNNER = {
        "bash": str(pipeline_runner_path),
        "powershell": f"& '{pipeline_runner_path}'",
    }
    executables = [
        mf.Executable(
            label="dream_3D_runner",
            instances=[
                mf.ExecutableInstance(
                    command=RUNNER[shell],
                    num_cores=1,
                    parallel_mode=None,
                ),
            ],
        ),
        mf.Executable(
            label="python_script",
            instances=[
                mf.ExecutableInstance(
                    command=(f'{__get_py_exe(shell)} "<<script_path>>" <<args>>'),
                    num_cores=1,
                    parallel_mode=None,
                ),
            ],
        ),
    ]
    setup = setup_runner + (mf.get_env_setup(shell) if use_current else setup_py)
    return mf.Environment(
        name="dream_3D_env",
        setup=setup,
        executables=executables,
        setup_label="dream3d",
    )


def env_configure_python(
    shell: Literal["bash", "powershell"],
    setup: str | list[str] | None = None,
    names: list[str] | None = None,
    use_current: bool = True,
) -> list[Environment]:
    """Configure Python MatFlow environments.

    Parameters
    ----------
    names:
        If specified, also set up these named environments using the same Python
        executable and setup, otherwise just set up the `python_env` environment.
    """
    setup = __norm_setup(setup)
    executables = [
        mf.Executable(
            label="python_script",
            instances=[
                mf.ExecutableInstance(
                    command=(f'{__get_py_exe(shell)} "<<script_path>>" <<args>>'),
                    num_cores=1,
                    parallel_mode=None,
                ),
            ],
        ),
    ]
    setup = mf.get_env_setup(shell) if use_current else setup
    environments = []
    for name in sorted(set(["python", *(names if names else [])])):
        environments.append(
            mf.Environment(
                name=f"{name}_env",
                setup=setup,
                executables=executables,
                setup_label="python",
            )
        )
    return environments


def env_configure_python_all(
    shell: Literal["bash", "powershell"],
    setup: str | list[str] | None = None,
    use_current: bool = True,
) -> list[Environment]:
    """
    Configure all of the Python environments, using the same setup.
    """
    return env_configure_python(
        shell=shell,
        setup=setup,
        use_current=use_current,
        names=(
            "damask_parse",
            "sklearn",
            "formable",
            "defdap",
            "gmsh",
            "moose_processing",
        ),
    )


def env_configure_matlab(
    shell: Literal["bash", "powershell"],
    setup: str | list[str] | None = None,
    matlab_path: str | None = None,
    matlab_runtime_path: str | None = None,
    mtex_path: str | None = None,
) -> Environment:
    """Configure the MATLAB MatFlow environment.

    Different environment executables are configured depending on what arguments are
    provided:

    1. If `matlab_path` and `mtex_path` are specified, then the `run_mtex`
       executable is configured.
    2. If `mcc_path` and `mtex_path` are specified, then the `compile_mtex` and
       `run_compiled_mtex` executables are configured.
    3. If `matlab_runtime_path` is specified, the `run_precompiled_mtex` executable is
       configured. Note that if `matlab_path` is specified, `matlab_runtime_path` will by
       default be set to the value of `matlab_path`.

    """

    matlab_exe = None
    matlab_mcc = None

    if matlab_path:
        matlab_path_ = Path(matlab_path)

        mcc_ext = ".bat" if shell == "powershell" else ""
        exe_ext = ".exe" if shell == "powershell" else ""

        matlab_exe = matlab_path_.joinpath("bin", f"matlab{exe_ext}")
        matlab_mcc = matlab_path_.joinpath("bin", f"mcc{mcc_ext}")

        matlab_runtime_path = matlab_runtime_path or matlab_path

        if shell != "powershell":
            matlab_exe = matlab_exe.as_posix()
            matlab_mcc = matlab_mcc.as_posix()

    run_mtex_cmd_nt = (
        f"& '{matlab_exe}' -batch \"addpath('<<script_dir>>'); "
        '<<script_name_no_ext>> <<args>>"'
    )
    run_mtex_cmd_posix = dedent(
        f"""\
        MTEX_DIR={mtex_path}
        for dir in $(find ${{MTEX_DIR}} -type d | grep -v -e ".git" -e "@" -e "private"); do MATLABPATH="${{dir}};${{MATLABPATH}}"; done
        export MATLABPATH=${{MATLABPATH}}
        {matlab_exe} -softwareopengl -singleCompThread -batch "addpath('<<script_dir>>'); <<script_name_no_ext>> <<args>>"
        """
    )
    compile_mtex_cmd_nt = (
        f"$mtex_path = '{mtex_path}'\n"
        f'& \'{matlab_mcc}\' -R -singleCompThread -m "<<script_path>>" <<args>> -o matlab_exe -a "$mtex_path/data" -a "$mtex_path/plotting/plotting_tools/colors.mat"'
    )

    compile_mtex_cmd_posix = dedent(
        f"""\
        MTEX_DIR={mtex_path}
        for dir in $(find ${{MTEX_DIR}} -type d | grep -v -e ".git" -e "@" -e "private" -e "data" -e "makeDoc" -e "templates" -e "nfft_openMP" -e "compatibility/")
        do
            MTEX_INCLUDE="-I ${{dir}} ${{MTEX_INCLUDE}}"
        done
        export MTEX_INCLUDE="${{MTEX_INCLUDE}} -a ${{MTEX_DIR}}/data -a ${{MTEX_DIR}}/plotting/plotting_tools/colors.mat"
        {matlab_mcc} -R -singleCompThread -R -softwareopengl -m "<<script_path>>" <<args>> -o matlab_exe ${{MTEX_INCLUDE}}
        """
    )

    run_compiled_mtex_nt = R".\matlab_exe.exe <<args>>"
    run_compiled_mtex_posix = dedent(
        f"""\
        export MATLAB_RUNTIME={matlab_runtime_path}
        ./run_matlab_exe.sh ${{MATLAB_RUNTIME}} <<args>>
        """
    )

    run_precompiled_mtex_nt = "& <<program_path>> <<args>>"
    run_precompiled_mtex_posix = dedent(
        f"""\
        export MATLAB_RUNTIME={matlab_runtime_path}
        <<program_path>> ${{MATLAB_RUNTIME}} <<args>>
        """
    )

    executables = []

    if matlab_path and mtex_path:

        mtex_path_ = Path(mtex_path)

        executables.append(
            mf.Executable(
                label="run_mtex",
                instances=[
                    mf.ExecutableInstance(
                        command=(
                            run_mtex_cmd_nt
                            if shell == "powershell"
                            else run_mtex_cmd_posix
                        ),
                        num_cores=1,
                        parallel_mode=None,
                    ),
                ],
            )
        )

        if matlab_mcc.is_file():
            executables.extend(
                [
                    mf.Executable(
                        label="compile_mtex",
                        instances=[
                            mf.ExecutableInstance(
                                command=(
                                    compile_mtex_cmd_nt
                                    if shell == "powershell"
                                    else compile_mtex_cmd_posix
                                ),
                                num_cores=1,
                                parallel_mode=None,
                            ),
                        ],
                    ),
                    mf.Executable(
                        label="run_compiled_mtex",
                        instances=[
                            mf.ExecutableInstance(
                                command=(
                                    run_compiled_mtex_nt
                                    if shell == "powershell"
                                    else run_compiled_mtex_posix
                                ),
                                num_cores=1,
                                parallel_mode=None,
                            ),
                        ],
                    ),
                ]
            )
        else:
            print(
                f"Not defining the `compile_mtex` executable because the MATLAB compiler "
                f"was not found."
            )

    if matlab_runtime_path:
        executables.append(
            mf.Executable(
                label="run_precompiled_mtex",
                instances=[
                    mf.ExecutableInstance(
                        command=(
                            run_precompiled_mtex_nt
                            if shell == "powershell"
                            else run_precompiled_mtex_posix
                        ),
                        num_cores=1,
                        parallel_mode=None,
                    ),
                ],
            )
        )

    new_env = mf.Environment(
        name="matlab_env",
        setup=setup,
        executables=executables,
        setup_label="matlab",
    )
    return new_env


def env_configure_damask(
    shell: Literal["bash", "powershell"],
    setup: str | list[str] | None = None,
    docker_image: str | None = None,
    docker_archive: str | Path | None = None,
    singularity_archive: str | Path | None = None,
    singularity_sif: str | Path | None = None,
    docker_exe: str = "docker",
    singularity_exe: str = "singularity",
):
    """Configure the MatFlow DAMASK environment.

    If passing a path to an archive, the name of the image within the archive can be
    passed via `docker_image`.

    Parameters
    ----------
    docker_image
        Name of the docker image to use.
    docker_archive:
        File path to an archived docker tar file. Use `docker_image` to set the image
        name within the tar (assumed to be "damask").
    singularity_archive
        File path to an archived docker tar file to be converted into a singularity
        sif file to use.
    singularity_sif
        File path to a Singularity sif file to use.
    """

    use_docker = bool(docker_image or docker_archive)
    use_singularity = bool(singularity_sif or singularity_archive)

    if use_docker and use_singularity:
        raise ValueError("Cannot use both singularity and docker!")

    if use_docker:
        if docker_image is None and docker_archive is None:
            raise ValueError(
                f"Provide either the docker image name, or the docker archive file path "
                f"(and optionally the image name within the archive)."
            )

        if docker_archive is not None and docker_image is None:
            # assume image name to be damask:
            docker_image = "damask"

        if docker_archive:
            # load with docker from an tar archive:
            cmd = (docker_exe, "load", "--input", str(docker_archive))
            run_cmd(cmd)

        if use_docker:
            DAMASK_GRID_CMD = {
                "bash": (
                    f"{docker_exe} run --rm --interactive --volume $PWD:/wd --env "
                    f"OMP_NUM_THREADS=$MATFLOW_RUN_NUM_THREADS {docker_image}"
                ),
                "powershell": (
                    f"{docker_exe} run --rm --interactive --volume ${{PWD}}:/wd --env "
                    f"OMP_NUM_THREADS=$MATFLOW_RUN_NUM_THREADS {docker_image}"
                ),
            }
    elif use_singularity:
        if singularity_archive:
            singularity_sif = singularity_sif or "damask.sif"  # i.e. pwd
            cmd = (
                singularity_exe,
                "build",
                singularity_sif,
                f"docker-archive://{singularity_archive}",
            )
            run_cmd(cmd)

        DAMASK_GRID_CMD = {
            "bash": (
                f"{singularity_exe} run -B $PWD:/wd {singularity_sif} --env "
                f"OMP_NUM_THREADS=$MATFLOW_RUN_NUM_THREADS"
            ),
        }

    executables = []
    executables.append(
        mf.Executable(
            label="damask_grid",
            instances=[
                mf.ExecutableInstance(
                    command=DAMASK_GRID_CMD[shell],
                    num_cores={"start": 1, "stop": 100, "step": 1},
                    parallel_mode=None,
                ),
            ],
        )
    )

    new_env = mf.Environment(
        name="damask_env",
        setup=setup,
        executables=executables,
        setup_label="damask",
    )
    return new_env


def env_configure_moose(
    shell: Literal["bash", "powershell"],
    setup: str | list[str] | None = None,
    docker_image: str | None = None,
    docker_archive: str | Path | None = None,
    singularity_archive: str | Path | None = None,
    singularity_sif: str | Path | None = None,
    docker_exe: str = "docker",
    singularity_exe: str = "singularity",
):
    """Configure the MatFlow MOOSE (proteus) environment.

    If passing a path to an archive, the name of the image within the archive can be
    passed via `docker_image`.

    Parameters
    ----------
    docker_image
        Name of the docker image to use.
    docker_archive:
        File path to an archived docker tar file. Use `docker_image` to set the image
        name within the tar (assumed to be "proteus").
    singularity_archive
        File path to an archived docker tar file to be converted into a singularity
        sif file to use.
    singularity_sif
        File path to a Singularity sif file to use.
    """

    use_docker = bool(docker_image or docker_archive)
    use_singularity = bool(singularity_sif or singularity_archive)

    if use_docker and use_singularity:
        raise ValueError("Cannot use both singularity and docker!")

    if use_docker:
        if docker_image is None and docker_archive is None:
            raise ValueError(
                f"Provide either the docker image name, or the docker archive file path "
                f"(and optionally the image name within the archive)."
            )

        if docker_archive is not None and docker_image is None:
            # assume image name to be proteus:
            docker_image = "proteus"

        if docker_archive:
            # load with docker from an tar archive:
            cmd = (docker_exe, "load", "--input", str(docker_archive))
            run_cmd(cmd)

        if use_docker:
            PROTEUS_CMD = {
                "bash": (
                    f"{docker_exe} run --rm --interactive --volume $PWD:/wd "
                    f"{docker_image} --n-threads=$MATFLOW_RUN_NUM_THREADS"
                ),
                "powershell": (
                    f"{docker_exe} run --rm --interactive --volume ${{PWD}}:/wd "
                    f"{docker_image} --n-threads=$MATFLOW_RUN_NUM_THREADS"
                ),
            }
    elif use_singularity:
        if singularity_archive:
            singularity_sif = singularity_sif or "proteus.sif"  # i.e. pwd
            cmd = (
                singularity_exe,
                "build",
                singularity_sif,
                f"docker-archive://{singularity_archive}",
            )
            run_cmd(cmd)

        PROTEUS_CMD = {
            "bash": (
                f"{singularity_exe} run -B $PWD:/wd {singularity_sif} "
                f"--n-threads=$MATFLOW_RUN_NUM_THREADS"
            ),
        }

    executables = []
    executables.append(
        mf.Executable(
            label="proteus",
            instances=[
                mf.ExecutableInstance(
                    command=PROTEUS_CMD[shell],
                    num_cores={"start": 1, "stop": 100, "step": 1},
                    parallel_mode=None,
                ),
            ],
        )
    )

    new_env = mf.Environment(
        name="moose_env",
        setup=setup,
        executables=executables,
        setup_label="moose",
    )
    return new_env



import numpy as np


def refine_plate_mesh_2D(SCF, hole_sect_nodes, plate_radial_nodes, plate_diff_nodes):

    # over all iterations:
    all_SCF = np.array(
        [SCF[k]["value"] for k in sorted(SCF, key=lambda x: int(x.split("_")[1]))]
    )
    print(f"refine_plate_mesh_2D: {all_SCF!r}")

    all_SCF_diff = np.diff(all_SCF)
    print(f"refine_plate_mesh_2D: {all_SCF_diff!r}")

    SCF_diff = all_SCF_diff[-1] if all_SCF_diff.size else 1e10
    print(f"refine_plate_mesh_2D: {SCF_diff!r}")

    scale = 1.2
    hole_sect_nodes = int(hole_sect_nodes * scale)
    plate_radial_nodes = int(plate_radial_nodes * scale)
    plate_diff_nodes = int(plate_diff_nodes * scale)

    # should be odd:
    if hole_sect_nodes % 2 == 0:
        hole_sect_nodes += 1
    if plate_diff_nodes % 2 == 0:
        plate_diff_nodes += 1

    return {
        "hole_sect_nodes": hole_sect_nodes,
        "plate_radial_nodes": plate_radial_nodes,
        "plate_diff_nodes": plate_diff_nodes,
        "SCF_diff": SCF_diff,
    }



def process_moose_parameters(p_perturbed, p_scales, index):
    # print(f"process_moose_parameters: {p_perturbed[:]=!r}")
    # print(f"process_moose_parameters: {p_scales[:]=!r}")

    E, nu = [i * j for i, j in zip(p_scales, p_perturbed[index])]
    # print(f"process_moose_parameters: {E=!r}; {nu=!r}")
    return {"elasticity": {"youngs_modulus": E, "poissons_ratio": nu}}



import pprint

import numpy as np


def increment_chain(x, g, all_x, all_g, all_accept, threshold):

    if g is None:
        # failed system analysis, reject the state:
        print(f"Increment chain: failed system analysis, rejecting state.")
        is_accept = False

    else:
        trial_x = x[:]  # convert to numpy array
        trial_g = g
        is_accept = trial_g > threshold

    current_x = all_x[-1]
    current_g = all_g[-1]

    new_x = trial_x if is_accept else current_x
    new_g = trial_g if is_accept else current_g

    all_x = np.vstack([all_x, new_x[None]])
    all_g = np.append(all_g, np.array(new_g))
    all_accept = np.append(all_accept, np.array(is_accept))

    return {
        "x": new_x,
        "g": new_g,
        "all_x": all_x,
        "all_g": all_g,
        "all_accept": all_accept,
    }



from pathlib import Path


_PLATE_WITH_HOLE = """\
//==============================================================================
// Gmsh 3D parametric plate with hole mesh
// author: Lloyd Fletcher (scepticalrabbit)
//==============================================================================
// Always set to OpenCASCADE - circles and boolean opts are much easier!
SetFactory("OpenCASCADE");

// Allows gmsh to print to terminal in vscode - easier debugging
General.Terminal = 1;

// View options - not required when
Geometry.PointLabels = 0;
Geometry.CurveLabels = 0;
Geometry.SurfaceLabels = 1;
Geometry.VolumeLabels = 0;

//------------------------------------------------------------------------------
// Variables
file_name = "plate_with_hole.msh";

// Geometric variables
plate_width = {plate_width};
plate_diff = {plate_diff}; // Must be positive
plate_height = plate_width+plate_diff;
plate_thick = {plate_thickness};

hole_rad = {hole_diameter}/2;
hole_loc_x = plate_width/2;
hole_loc_y = plate_height/2;
hole_circ = 2*Pi*hole_rad;

// Mesh variables
elem_order = 1;

plate_thick_layers = {plate_thick_layers};
hole_sect_nodes = {hole_sect_nodes}; // Must be odd
plate_rad_nodes = {plate_radial_nodes};
plate_diff_nodes = {plate_diff_nodes}; // numbers of nodes along the rectangular extension
plate_edge_nodes = Floor((hole_sect_nodes-1)/2)+1;
elem_size = hole_circ/(4*(hole_sect_nodes-1));
tol = elem_size; // Used for bounding box selection tolerance
tol_thick = plate_thick/(4*plate_thick_layers);

//------------------------------------------------------------------------------
// Geometry Definition

// Split plate into eight pieces with a square around the hole to allow spider
// web meshing around the hole
s1 = news;
Rectangle(s1) = {{0.0,0.0,0.0,
                plate_width/2,plate_diff/2}};
s2 = news;
Rectangle(s2) = {{plate_width/2,0.0,0.0,
                plate_width/2,plate_diff/2}};

s3 = news;
Rectangle(s3) = {{0.0,plate_diff/2,0.0,
                plate_width/2,plate_width/2}};
s4 = news;
Rectangle(s4) = {{plate_width/2,plate_diff/2,0.0,
                plate_width/2,plate_width/2}};

s5 = news;
Rectangle(s5) = {{0.0,plate_width/2+plate_diff/2,0.0,
                plate_width/2,plate_width/2}};
s6 = news;
Rectangle(s6) = {{plate_width/2,plate_width/2+plate_diff/2,0.0,
                plate_width/2,plate_width/2}};

s7 = news;
Rectangle(s7) = {{0.0,plate_height-plate_diff/2,0.0,
                plate_width/2,plate_diff/2}};
s8 = news;
Rectangle(s8) = {{plate_width/2,plate_height-plate_diff/2,0.0,
                plate_width/2,plate_diff/2}};

// Merge coincicent edges of the four overlapping squares
BooleanFragments{{ Surface{{s1}}; Delete; }}
                {{ Surface{{s2,s3,s4,s5,s6,s7,s8}}; Delete; }}

// Create the hole surface
c2 = newc; Circle(c2) = {{hole_loc_x,hole_loc_y,0.0,hole_rad}};
cl2 = newcl; Curve Loop(cl2) = {{c2}};
s9 = news; Plane Surface(s9) = {{cl2}};
// Bore out the hole from the quarters of the plate
BooleanDifference{{ Surface{{s3,s4,s5,s6}}; Delete; }}{{ Surface{{s9}}; Delete; }}

//------------------------------------------------------------------------------
// Transfinite meshing (line element sizes and mapped meshing)
Transfinite Curve{{31,24,26,28}} = plate_rad_nodes;
Transfinite Curve{{1,5,3,7,23,29,30,34,14,17,19,22}} = plate_edge_nodes;
Transfinite Curve{{32,33,25,27}} = hole_sect_nodes;
Transfinite Curve{{4,2,6,20,18,21}} = plate_diff_nodes;

// NOTE: recombine surface turns default triangles into quads

Transfinite Surface{{s1}} = {{1,2,3,4}};
Recombine Surface{{s1}};
Transfinite Surface{{s2}} = {{2,5,6,3}};
Recombine Surface{{s2}};

Transfinite Surface{{s3}} = {{17,18,3,16}};
Recombine Surface{{s3}};
Transfinite Surface{{s4}} = {{18,19,20,3}};
Recombine Surface{{s4}};
Transfinite Surface{{s5}} = {{17,21,10,16}};
Recombine Surface{{s5}};
Transfinite Surface{{s6}} = {{19,21,10,20}};
Recombine Surface{{s6}};

Transfinite Surface{{s7}} = {{11,10,13,14}};
Recombine Surface{{s7}};
Transfinite Surface{{s8}} = {{10,12,15,13}};
Recombine Surface{{s8}};

Extrude{{0.0,0.0,plate_thick}}{{
    Surface{{:}}; Layers{{plate_thick_layers}}; Recombine;
}}

//------------------------------------------------------------------------------
// Physical lines and surfaces for export/BCs
Physical Volume("plate-vol") = {{Volume{{:}}}};

ps1() = Surface In BoundingBox{{
    0.0-tol,plate_height-tol,0.0-tol,
    plate_width+tol,plate_height+tol,plate_thick+tol}};
Physical Surface("bc-top-disp") = {{ps1(0),ps1(1)}};

ps2() = Surface In BoundingBox{{
    0.0-tol,0.0-tol,0.0-tol,
    plate_width+tol,0.0+tol,plate_thick+tol}};
Physical Surface("bc-base-disp") = {{ps2(0),ps2(1)}};

ps3() = Surface In BoundingBox{{
    0.0-tol,0.0-tol,plate_thick-tol_thick,
    plate_width+tol,plate_height+tol,plate_thick+tol_thick}};
Physical Surface("plate-surf-vis-front") = {{ps3(0),ps3(1),ps3(2),ps3(3),
                                            ps3(4),ps3(5),ps3(6),ps3(7)}};

ps4() = Surface In BoundingBox{{
    0.0-tol,0.0-tol,0.0-tol_thick,
    plate_width+tol,plate_height+tol,0.0+tol_thick}};
Physical Surface("plate-surf-vis-back") = {{ps4(0),ps4(1),ps4(2),ps4(3),
                                            ps4(4),ps4(5),ps4(6),ps4(7)}};

//------------------------------------------------------------------------------
// Global meshing
num_threads = 4;

Mesh.Algorithm = 6;
Mesh.Algorithm3D = 10;

General.NumThreads = num_threads;
Mesh.MaxNumThreads1D = num_threads;
Mesh.MaxNumThreads2D = num_threads;
Mesh.MaxNumThreads3D = num_threads;

Mesh.ElementOrder = elem_order;
Mesh 3;

//------------------------------------------------------------------------------
// Save and exit
Save Str(file_name);
Exit;
"""

_LOOKUP = {
    "plate_with_hole": _PLATE_WITH_HOLE,
}


def generate_geo(path, geometry_name, **kwargs):
    """Generate a Gmsh .geo file for generating a mesh of a plate with a hole."""
    with Path(path).open("wt") as fh:
        fh.write(_LOOKUP[geometry_name.lower()].format(**kwargs))



import numpy as np


def model_exp(x, *p):
    a, b, c = p
    return a * np.exp(b * x) + c


def get_reference_data_toy_model(p_true):
    # print(f"get_reference_data_toy_model: {p_true=!r}")
    np.random.seed(0)
    x_data = np.linspace(0, 2, 30)
    y_clean = model_exp(x_data, *p_true)
    y_ref = y_clean + 0.12 * np.random.randn(x_data.size)  # add noise
    return {"x": x_data, "reference_response": y_ref}



import numpy as np
from scipy.stats import norm, uniform
from damask import Grid
from damask_parse.utils import validate_volume_element, validate_orientations


def quat_sample_random(rand_nums):
    """Generate random uniformly distributed unit quaternions.

    Parameters
    ----------
    rand_nums : ndarray of shape (N, 3)

    Returns
    -------
    quats : ndarray of shape (N, 4)

    References
    ----------
    https://stackoverflow.com/a/44031492/5042280
    https://web.archive.org/web/20220517133017/http://planning.cs.uiuc.edu/node198.html

    """
    return np.array(
        [
            np.sqrt(1 - rand_nums[:, 0]) * np.sin(2 * np.pi * rand_nums[:, 1]),
            np.sqrt(1 - rand_nums[:, 0]) * np.cos(2 * np.pi * rand_nums[:, 1]),
            np.sqrt(rand_nums[:, 0]) * np.sin(2 * np.pi * rand_nums[:, 2]),
            np.sqrt(rand_nums[:, 0]) * np.cos(2 * np.pi * rand_nums[:, 2]),
        ]
    ).T


def _to_standard_normal_space(dist, variates):
    """Convert random variables defined in some non-standard-normal continuous
    distribution space to a standard normal distribution space."""
    return norm.ppf(dist.cdf(variates), loc=0, scale=1)


def _to_physical_space(dist, variates):
    """Convert random variables defined in standard normal space to some other
    continuous distribution space."""
    return dist.ppf(norm.cdf(variates, loc=0, scale=1))


def standard_normal_to_uniform_space(dist, variates):
    min_, max_ = dist.support()
    return norm.cdf(variates, loc=0, scale=1) * (max_ - min_) + min_


def uniform_to_standard_normal_space(dist, variates):
    min_, max_ = dist.support()
    return norm.ppf(((variates - min_) / (max_ - min_)), loc=0, scale=1)


def to_physical_space(dists, variates):
    """Convert random variables sampled from the standard normal space to the specified
    distribution spaces.

    Parameters
    ----------
    dists
        List of SciPy frozen distribution instances
    variates
        Standard-normally distributed random variates. Outer dimension must
        match length of `dists` list.
    """
    funcs = [_lookup.get(i.dist.name, ("", _to_physical_space))[1] for i in dists]
    assert len(funcs) == len(variates)
    inputs_t = np.empty_like(variates)
    for idx, (func_i, ins_i) in enumerate(zip(funcs, variates)):
        inputs_t[idx] = func_i(dists[idx], ins_i)
    return inputs_t


_lookup = {
    "uniform": (uniform_to_standard_normal_space, standard_normal_to_uniform_space),
}


def generate_volume_element_voronoi_variates(
    x,
    VE_grid_size,
    VE_size,
    phase_label,
    homog_label,
    scale_morphology,
    scale_update_size,
):

    random_variates = x
    print(f"generate RVE: {random_variates.shape=}")

    num_vars = len(x)
    inputs = [uniform() for _ in range(num_vars)]
    uniform_vars = to_physical_space(inputs, random_variates)

    num_grains_ = num_vars / 6  # three for seed point (x, y, z), three for quaternion
    assert num_grains_ % 1 == 0
    num_grains = int(num_grains_)

    seeds = uniform_vars[: 3 * num_grains].reshape((num_grains, 3))
    ori_numbers = uniform_vars[3 * num_grains :].reshape((num_grains, 3))
    quats = quat_sample_random(ori_numbers)

    grid_obj = Grid.from_Voronoi_tessellation(
        cells=np.array(VE_grid_size),
        size=np.array(VE_size),
        seeds=seeds,
        periodic=True,
    )

    if scale_morphology is not None:
        scale_morphology = np.array(scale_morphology)

        original_cells = grid_obj.cells
        new_cells = original_cells * scale_morphology
        grid_scaled = grid_obj.scale(new_cells)

        if scale_update_size:
            original_size = grid_obj.size
            new_size = original_size * scale_morphology
            grid_scaled.size = new_size

        grid_obj = grid_scaled

    unit_cell_alignment = {
        "x": "a",
        "y": "b*",
        "z": "c",
    }
    oris = {
        "type": "quat",
        "unit_cell_alignment": unit_cell_alignment,
        "quaternions": quats,
        "quat_component_ordering": "scalar-vector",
    }

    oris = validate_orientations(oris)

    const_phase_lab = np.array([phase_label])[np.zeros(num_grains, dtype=int)]
    ori_idx = np.arange(num_grains)

    volume_element = {
        "size": grid_obj.size.astype(float).tolist(),
        "grid_size": grid_obj.cells.tolist(),
        "orientations": oris,
        "element_material_idx": grid_obj.material,
        "constituent_material_idx": np.arange(num_grains),
        "constituent_material_fraction": np.ones(num_grains),
        "constituent_phase_label": const_phase_lab,
        "constituent_orientation_idx": ori_idx,
        "material_homog": np.full(num_grains, homog_label),
    }
    return {"volume_element": validate_volume_element(volume_element)}



from __future__ import annotations
from pathlib import Path
from damask_parse.readers import read_HDF5_file


def extract_damask_HDF5(
    damask_hdf5_file: Path | str, VE_response_data: dict[str, list[dict]]
) -> dict:
    """Operate on and extract data from an HDF5 file generated by a DAMASK run.

    Parameters
    ----------
    damask_hdf5_path : Path or str
        Path to the HDF5 file.
    VE_response_data : dict
        Contents are as follows:

        volume_data : list of dict, optional
            List of data to extract from the entire volume. This is a list of dict
            with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
                out_name: str, optional
                    Name of the data
                transforms: list of dict, optional
                    If specified this is a list of dicts, each with the following keys:
                        sum_along_axes : int, optional
                            If specified, take the sum the array along this axis.
                        mean_along_axes: int, optional
                            If specified, take the mean average of the array along this
                            axis.

        phase_data : list of dict, optional
            List of data to extract from a single phase. This is a list of dict
            with following keys:
                field_name: str
                    Name of the data field to extract
                phase_name : str
                    Name of phase to
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
                out_name: str, optional
                    Name of the data
                transforms: list of dict, optional
                    If specified this is a list of dicts, each with the following keys:
                        sum_along_axes : int, optional
                            If specified, take the sum the array along this axis.
                        mean_along_axes: int, optional
                            If specified, take the mean average of the array along this
                            axis.

        field_data : list of dict, optional
            List of field data to extract. Only extracts for the first constituent
            of each material point. This is a list of dict with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
            Special field_name keys exist, 'grain' and 'phase'. Use 'u_n' or 'u_p'
            for displacement.

        grain_data : list of dict, optional
            List of grain data to extract. Only extracts for the first constituent
            of each material point. This is a list of dict with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract

    Returns
    -------
    volume_element_response : dict
        Dict with keys determined by the `field_data` dict.
    """
    VE_response = read_HDF5_file(hdf5_path=damask_hdf5_file, **VE_response_data)
    return VE_response



import logging

import numpy as np
from numpy.typing import NDArray
from scipy.stats import norm


def set_up_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(level=logging.DEBUG)
    fh = logging.FileHandler("script.log")
    fh_formatter = logging.Formatter(
        "%(asctime)s %(levelname)s:%(filename)s: %(message)s"
    )
    fh.setFormatter(fh_formatter)
    logger.addHandler(fh)
    return logger


def generate_next_state(x, prop_std):
    """Generate the next candidate state in a modified Metropolis algorithm.

    Parameters
    ----------
    x
        Current state on which the candidate state will depend.

    Returns
    -------
    dict:
        x:
            Generated candidate state.
    """

    x = x[:]  # convert to numpy array
    dim = len(x)
    current_state = x
    rng = np.random.default_rng()
    xi = np.empty(dim)

    proposal = norm(loc=current_state, scale=prop_std)
    xi_hat = np.atleast_1d(proposal.rvs())
    accept_ratios = np.divide(*norm.pdf([xi_hat, current_state]))
    accept_idx = rng.random(len(accept_ratios)) < np.minimum(1, accept_ratios)

    xi[accept_idx] = xi_hat[accept_idx]
    xi[~accept_idx] = current_state[~accept_idx]

    return {"x": xi}



from __future__ import annotations
from pathlib import Path
from damask_parse.readers import read_HDF5_file
from damask_parse.utils import generate_viz


def process_HDF5(
    damask_hdf5_file: Path | str,
    damask_post_processing: list[dict],
    VE_response_data: dict,
    damask_viz: list[dict] | dict | bool | None,
    remove_damask_hdf5: bool,
):
    """
    Operate on and extract data from an HDF5 file generated by a DAMASK run.

    Supported post-processing operations
    ------------------------------------
    - `add_absolute`
    - `add_calculation`
    - `add_stress_Cauchy`
    - `add_determinant`
    - `add_deviator`
    - `add_eigenvalue`
    - `add_eigenvector`
    - `add_IPF_color`
    - `add_maximum_shear`
    - `add_equivalent_Mises`
    - `add_norm`
    - `add_stress_second_Piola_Kirchhoff`
    - `add_pole`
    - `add_rotation`
    - `add_spherical`
    - `add_strain`
    - `add_stretch_tensor`
    - `add_curl`
    - `add_divergence`
    - `add_gradient`

    Parameters
    ----------
    damask_hdf5_path : Path or str
        Path to the HDF5 file.
    damask_post_processing
        List of methods to invoke on the DADF5 object. This is a list of dicts with the
        following keys:
            name : str
                The name of the DADF5 method.
            args : dict
                Parameter names and their values to pass to the DADF5 method. This
                assumes all DADF5 method parameters are of positional-or-keyword type.
            opts : dict, optional
                Additional options.
    VE_response_data : dict
        Contents are as follows:

        volume_data : list of dict, optional
            List of data to extract from the entire volume. This is a list of dict
            with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
                out_name: str, optional
                    Name of the data
                transforms: list of dict, optional
                    If specified this is a list of dicts, each with the following keys:
                        sum_along_axes : int, optional
                            If specified, take the sum the array along this axis.
                        mean_along_axes: int, optional
                            If specified, take the mean average of the array along this
                            axis.

        phase_data : list of dict, optional
            List of data to extract from a single phase. This is a list of dict
            with following keys:
                field_name: str
                    Name of the data field to extract
                phase_name : str
                    Name of phase to
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
                out_name: str, optional
                    Name of the data
                transforms: list of dict, optional
                    If specified this is a list of dicts, each with the following keys:
                        sum_along_axes : int, optional
                            If specified, take the sum the array along this axis.
                        mean_along_axes: int, optional
                            If specified, take the mean average of the array along this
                            axis.

        field_data : list of dict, optional
            List of field data to extract. Only extracts for the first constituent
            of each material point. This is a list of dict with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
            Special field_name keys exist, 'grain' and 'phase'. Use 'u_n' or 'u_p'
            for displacement.

        grain_data : list of dict, optional
            List of grain data to extract. Only extracts for the first constituent
            of each material point. This is a list of dict with following keys:
                field_name: str
                    Name of the data field to extract
                increments: list of dict, optional
                    List of increment specifications to extract data from. Values
                    refer to increments in the simulation. Default to all. This is
                    a list of dict one of the following sets of keys:
                        values: list of int
                            List of incremnts to extract
                        ----OR----
                        start: int
                            First increment to extract
                        stop: int
                            Final incremnt to extract (inclusive)
                        step: int
                            Step between increments to extract
    damask_viz
        Visualisation control data.

    Returns
    -------
    volume_element_response : dict
        Dict with keys determined by the `field_data` dict.
    """
    VE_response = read_HDF5_file(
        hdf5_path=damask_hdf5_file,
        operations=damask_post_processing,
        **VE_response_data,
    )
    generate_viz(hdf5_path=damask_hdf5_file, viz_spec=damask_viz, parsed_outs=VE_response)

    # remove damask HDF5 file (TODO: needs to be a better way to do this in hpcflow)
    if remove_damask_hdf5:
        damask_hdf5_file.unlink()

    return VE_response



"""
Surrogate model building, fitting, and Bayesian inference with Scikit-learn and PyMC.

Note: some methods of the Surrogate class require additional dependencies that are not
included by MatFlow; namely: sklearn, pytensor, pymc, and on Windows, sklearnex.

This code is taken entirely (with minor modifications) from:
https://github.com/LightForm-group/surrogatetools.

Some imports (including skops.io, scipy.stats, and scipy.optimize) are inlined into the
methods where they are needed to reduce the effect on MatFlow's import time.

"""

from __future__ import annotations
from typing import ClassVar
import numpy as np
from hpcflow.sdk.core.parameters import ParameterValue

import matflow as mf


class Surrogate(ParameterValue):

    _typ: ClassVar[str] = "surrogate"

    #: Whether we have yet patched scikit-learn with the Intel extension package.
    _is_patched: ClassVar[bool] = False

    #: Path to the c++ compiler that Pytensor should use. If setting, this should be done
    #: before an instance is created.
    pytensor_config_cxx: ClassVar[str | None] = None

    #: Whether we have yet configured Pytensor's cxx config option.
    _pytensor_config_cxx_is_set: ClassVar[bool] = False

    def __init__(self, X, y, parameter_names=None, model=None, scale=True):

        self._patch_once()

        self.X_scaler = None
        self.y_scaler = None

        self.X = X
        self.y = y

        self.parameter_range = np.array(
            [np.min(self.X, axis=0), np.max(self.X, axis=0)]
        ).T

        self.parameter_names = parameter_names

        self.N, self.D = self.X.shape
        self.N, self.P = self.y.shape

        if model:
            trusted = [
                "sklearn.gaussian_process.kernels.ConstantKernel",
                "sklearn.gaussian_process.kernels.Matern",
                "sklearn.gaussian_process.kernels.Product",
                "sklearn.gaussian_process.kernels.Sum",
            ]
            import skops.io as sio

            self.model = sio.loads(model, trusted=trusted)
        else:
            self.model = None

        self._scale = scale
        if scale:
            self.scale_data()

    def _serialise_model(self) -> bytes:
        """Serialise the scikit-learn model to bytes using skops."""

        if model := self.model:
            import skops.io as sio

            return sio.dumps(model)

        else:
            return None

    def to_dict(self):
        """Convert to a serialisable dictionary."""
        X, y = self.get_unscaled_data() if self._scale else (self.X, self.y)
        return {
            "X": X,
            "y": y,
            "parameter_names": self.parameter_names,
            "model": self._serialise_model(),
            "scale": self._scale,
        }

    def scale_data(self, scale_X=True, scale_y=True):

        from sklearn.preprocessing import RobustScaler

        if self.X_scaler == None:
            self.X_scaler = RobustScaler().fit(self.X)

        if self.y_scaler == None:
            self.y_scaler = RobustScaler().fit(self.y)

        if scale_X == True:
            self.X = self.X_scaler.transform(self.X)

        if scale_y == True:
            self.y = self.y_scaler.transform(self.y)

        return None

    def get_unscaled_data(self, unscale_X=True, unscale_y=True):

        un_X = None
        if unscale_X:
            un_X = self.X_scaler.inverse_transform(self.X)

        un_y = None
        if unscale_y:
            un_y = self.y_scaler.inverse_transform(self.y)

        return un_X, un_y

    @classmethod
    def _patch_once(cls):
        if not cls._is_patched:
            if mf.run_time_info.platform != "macos":
                from sklearnex import patch_sklearn

                patch_sklearn()
            cls._is_patched = True

    def build_model(
        self,
        kernel="matern",
        cross_validate=False,
        cross_validator=5,
        scoring="neg_mean_absolute_error",
        **kwargs,
    ):

        from sklearn.gaussian_process import GaussianProcessRegressor
        from sklearn.gaussian_process.kernels import Matern
        from sklearn.model_selection import cross_val_score

        # Firstly we define the kernel to be used
        if kernel == "matern":
            kernel = 1.0 * Matern(length_scale=self.D * [1], nu=2.5) + 1.0

        else:
            kernel = kernel

        # Build the kernel along with any additional parameters

        # Optimise with cross validation
        if cross_validate == True:

            model = GaussianProcessRegressor(kernel=kernel, **kwargs)
            score = cross_val_score(
                model, self.X, self.y, cv=cross_validator, scoring=scoring
            )

            print("Cross validation score: ", np.round(score, 2))

        self.model = GaussianProcessRegressor(kernel=kernel, **kwargs)
        self.model.fit(self.X, self.y)

    def make_prediction(self, X, return_std=False, scalar_output=False):

        if self.X_scaler is not None:
            X = self.X_scaler.transform(X)

        y_prediction, y_error = self.model.predict(X, return_std=True)

        if scalar_output == True:
            y_prediction = y_prediction.reshape(1, -1)

        if self.y_scaler is not None:
            y_prediction = self.y_scaler.inverse_transform(y_prediction)
            y_error = y_error * self.y_scaler.scale_

        if return_std:
            return y_prediction, y_error
        else:
            return y_prediction

    def make_prediction_sobol(self, X):

        X = np.array(X).T

        y_prediction = self.make_prediction(X, return_std=False)

        return y_prediction.T

    def generate_sobol(self, use_fit=False, **kwargs):
        """
        Wrapper function for scipy.stats.sobol_indices

        Args:
            n (int): number of samples

            **kwargs: additonal parameters

        Returns:
            sobol: SobolResult
        """

        from scipy.stats import norm, uniform, sobol_indices

        if use_fit == True:
            distributions = [
                norm(loc=self.parameter_mean[i], scale=self.parameter_std[i])
                for i in range(len(self.parameter_range))
            ]

        else:
            distributions = [
                uniform(loc=x[0], scale=x[1] - x[0]) for x in self.parameter_range
            ]

        sobol = sobol_indices(
            func=self.make_prediction_sobol, dists=distributions, **kwargs
        )

        return sobol

    def fit(self, Y_actual, Y_error, use_std=True, **kwargs):
        from scipy.optimize import shgo

        def _loss(params, Y_actual, Y_error):

            params = np.array(params).reshape(1, -1)

            if use_std == True:

                y_prediction, y_prediction_error = self.make_prediction(
                    X=params, return_std=True
                )

                residual_square = (
                    y_prediction[0] - Y_actual
                ) ** 2 + y_prediction_error**2

            else:
                y_prediction = self.make_prediction(X=params)

                residual_square = (y_prediction[0] - Y_actual) ** 2

            error_residual_square = Y_error**2

            loss = residual_square / error_residual_square

            return loss.sum()

        res = shgo(_loss, bounds=self.parameter_range, args=(Y_actual, Y_error), **kwargs)

        return res

    def _config_pytensor_once(self):
        """
        Configure the path to the C++ compiler that pytensor should use, using the
        class variable `pytensor_config_cxx` if set, or, if on Apple silicon, setting to a
        sensible default.
        """

        if self._pytensor_config_cxx_is_set:
            return

        import pytensor

        if (pt_cxx := self.pytensor_config_cxx) is not None:
            pytensor.config.cxx = pt_cxx
        elif mf.run_time_info.is_apple_silicon:
            pytensor.config.cxx = "/usr/bin/clang++"

        self._pytensor_config_cxx_is_set = True

    def perform_inference(self, Y_actual, Y_error, initval=None, use_std=True, **kwargs):

        self._config_pytensor_once()
        import pymc as pm
        import pytensor.tensor as pt
        from pytensor.graph import Apply, Op

        data = [Y_actual, Y_error]

        def my_loglike(params, data):
            # We fail explicitly if inputs are not numerical types for the sake of this tutorial
            # As defined, my_loglike would actually work fine with PyTensor variables!
            params = np.array(params).reshape(1, -1)

            y_actual, y_actual_error = data

            # The surrogate must provide a prediction and uncertainty estimate
            if use_std == True:
                y_prediction, y_prediction_error = self.make_prediction(
                    params, return_std=True
                )

                residual_square = (y_prediction[0] - y_actual) ** 2 + y_prediction_error[
                    0
                ] ** 2

            else:
                y_prediction = self.make_prediction(params)

                residual_square = (y_prediction[0] - y_actual) ** 2

            loss = residual_square / y_actual_error**2

            f = -0.5 * loss

            return f

        class LogLike(Op):
            def make_node(self, params, data) -> Apply:
                # Convert inputs to tensor variables
                params = pt.as_tensor(params)
                data = pt.as_tensor(data)

                inputs = [params, data]
                # Define output type, in our case a vector of likelihoods
                # with the same dimensions and same data type as data
                # If data must always be a vector, we could have hard-coded
                outputs = [pt.vector()]

                # Apply is an object that combines inputs, outputs and an Op (self)
                return Apply(self, inputs, outputs)

            def perform(
                self, node: Apply, inputs: list[np.ndarray], outputs: list[list[None]]
            ) -> None:
                # This is the method that compute numerical output
                # given numerical inputs. Everything here is numpy arrays
                params, data = inputs  # this will contain my variables

                # call our numpy log-likelihood function
                loglike_eval = my_loglike(params, data)

                # Save the result in the outputs list provided by PyTensor
                # There is one list per output, each containing another list
                # pre-populated with a `None` where the result should be saved.
                outputs[0][0] = np.asarray(loglike_eval)

        loglike_op = LogLike()

        def custom_dist_loglike(data, params):

            # create our Op
            # data, or observed is always passed as the first input of CustomDist
            return loglike_op(params, data)

        # use PyMC to sampler from log-likelihood
        with pm.Model() as no_grad_model:

            params = []

            for i in range(self.D):

                distribution = pm.Uniform(
                    self.parameter_names[i],
                    lower=self.parameter_range[i][0],
                    upper=self.parameter_range[i][1],
                    initval=initval[i],
                )

                params.append(distribution)

            # use a CustomDist with a custom logp function
            likelihood = pm.CustomDist(
                "likelihood",
                params,
                observed=data,
                logp=custom_dist_loglike,
            )

        ip = no_grad_model.initial_point()

        no_grad_model.compile_logp(vars=[likelihood], sum=False)(ip)

        with no_grad_model:

            step = pm.DEMetropolisZ()

            # Use custom number of draws to replace the HMC based defaults
            idata_no_grad = pm.sample(
                step=step, **kwargs
            )  # 50_000, tune=50_000,cores=4,chains=4,step=step,return_inferencedata=True)

        return idata_no_grad
